
#ifndef PLS_DATAFLOW_INTERNAL_GRAPH_H_
#define PLS_DATAFLOW_INTERNAL_GRAPH_H_

#include <atomic>
#include <tuple>
#include <iostream>

#include "node.h"
#include "in_port.h"
#include "out_port.h"

#include "pls/dataflow/inputs.h"
#include "pls/dataflow/outputs.h"

namespace pls {
namespace dataflow {
namespace internal {

template<typename INS, typename OUTS>
class graph {};

template<typename I0, typename ...I, typename O0, typename ...O>
class graph<pls::dataflow::inputs<I0, I...>, pls::dataflow::outputs<O0, O...>> : public node {

 private:
  // Our own type
  using self_type = graph<pls::dataflow::inputs<I0, I...>, pls::dataflow::outputs<O0, O...>>;

  // Input-Output port types
  using multi_in_port_type = multi_in_port<self_type, 0, I0, I...>;
  using multi_out_port_type = multi_out_port<O0, O...>;

  // Input-Output value tuples
  using input_tuple = std::tuple<token<I0>, token<I>...>;
  using output_tuple = std::tuple<token<O0>, token<O>...>;
  static constexpr int num_in_ports = std::tuple_size<input_tuple>();
  static constexpr int num_out_ports = std::tuple_size<output_tuple>();

  // Memory used for ONE active invocation of the dataflow-graph
  struct invocation_memory {
    std::atomic<int> inputs_missing_;
    input_tuple input_buffer_;
  };

 public:
  template<int POS>
  using in_port_at = typename multi_in_port_type::template in_port_type_at<POS>;
  template<int POS>
  using out_port_at = typename multi_out_port_type::template out_port_type_at<POS>;

  template<int POS>
  in_port_at<POS> &in_port() {
    return in_port_.template get<POS>();
  }

  template<int POS>
  out_port_at<POS> &out_port() {
    return out_port_.template get<POS>();
  }

  template<int POS, typename T>
  void token_pushed(token<T> token) {
    auto invocation = get_invocation<invocation_memory>(token);

    std::cout << "Token pushed at " << POS << " with value " << token.value() << std::endl;
    auto remaining_inputs = --(invocation->inputs_missing_);
    if (remaining_inputs == 0) {
      std::cout << "All inputs there!" << std::endl;
    }
  }

  graph() : in_port_{this, this}, out_port_{} {
  }

  /////////////////////////////////////////////////////////////////
  // Graph building
  /////////////////////////////////////////////////////////////////
  void build() {
    state_ = building;
    PLS_ASSERT(is_fully_connected(), "Must fully connect dataflow graphs!")

    num_nodes_ = 0;
    for (int i = 0; i < num_in_ports; i++) {
      build_recursive(successor_at(i));
    }

    state_ = built;
  }

  node *start_{nullptr};
  node *current_{nullptr};

  void build_recursive(node *node) {
    if (node->state_ != fresh) {
      return; // Already visited
    }
    node->state_ = building;
    PLS_ASSERT(node->is_fully_connected(), "Must fully connect dataflow graphs")

    add_node(node);
    for (int i = 0; i < node->num_successors(); i++) {
      build_recursive(node->successor_at(i));
    }

    node->state_ = built;
  }

  void add_node(node *new_node) {
    new_node->memory_index_ = num_nodes_++;

    if (start_ == nullptr) {
      start_ = new_node;
      current_ = new_node;
    } else {
      current_->direct_successor_ = new_node;
      current_ = new_node;
    }
  }

  //////////////////////////////////////////////////////////////////
  // Overrides for generic node functionality (building the graph)
  //////////////////////////////////////////////////////////////////
  int num_successors() const override {
    return num_out_ports;
  }
  node *successor_at(int pos) const override {
    return out_port_.next_node_at(pos);
  }

  bool is_fully_connected() const override {
    return in_port_.fully_connected() && out_port_.fully_connected();
  }

  int instance_buffer_size() const override {
    return sizeof(invocation_memory);
  }
  void init_instance_buffer(void *memory) const override {
    auto invocation = new(memory) invocation_memory{};
    invocation->inputs_missing_ = num_in_ports;
  };

 private:
  // Input/Output ports
  multi_in_port_type in_port_;
  multi_out_port_type out_port_;

  int num_nodes_{0};
};

}
}
}

#endif //PLS_DATAFLOW_INTERNAL_GRAPH_H_
